import torch

def tuple_norm_square(tuple_in):
    norm_in = 0.
    for t in tuple_in:
        norm_in += t.square().sum()
    return norm_in

def zo_grad_forward(closure, params, tau, *args, **kwargs):
    with torch.no_grad():
        e_vector = [torch.randn_like(y) for y in params]
        norm = tuple_norm_square(e_vector).sqrt()
        f_0 = closure()
        d = 0
        for y, e in zip(params, e_vector):
            d += torch.numel(y)
            e.div_(norm)
            y.add_(e.mul(tau))
        f_1 = closure()
        for y, e in zip(params, e_vector):
            y.sub_(e.mul(tau))
        scaler = (f_1 - f_0).mul(d / tau)
        return [e.mul(scaler) for e in e_vector]


def zo_grad_forward_sample(closure, params, tau, *args, sample_size=1, **kwargs):
    with torch.no_grad():
        params_0 = [y.clone() for y in params]
        if sample_size == 1:
            inexact_gradient = zo_grad_forward(closure, params, tau)
        else:
            inexact_gradient = [torch.zeros_like(y) for y in params]
            f_0 = closure()
            for _ in range(sample_size):
                e_vector = [torch.randn_like(y) for y in params]
                norm = tuple_norm_square(e_vector).sqrt()
                d = 0
                for y, e in zip(params, e_vector):
                    d += torch.numel(y)
                    e.div_(norm)
                    y.add_(e.mul(tau))
                f_1 = closure()
                for y, e in zip(params, e_vector):
                    y.sub_(e.mul(tau))
                scaler = (f_1 - f_0).mul(d / tau)
                for e, g in zip(e_vector, inexact_gradient):
                    g += e.mul(scaler)
        for y, p in zip(params, params_0):
            y.zero_().add_(p)
    return inexact_gradient


def zo_grad_central(closure, params, tau, *args, **kwargs):
    with torch.no_grad():
        e_vector = [torch.randn_like(y) for y in params]
        norm = tuple_norm_square(e_vector).sqrt()
        d = 0
        for y, e in zip(params, e_vector):
            d += torch.numel(y)
            e.div_(norm)
            y.add_(e.mul(tau))

        f_1 = closure()
        for y, e in zip(params, e_vector):
            y.sub_(e.mul(2 * tau))
        f_0 = closure()
        for y, e in zip(params, e_vector):
            y.add_(e.mul(tau))

        scaler = (f_1 - f_0).mul(d / (2 * tau))
    return [e.mul(scaler) for e in e_vector]


def zo_grad_central_sample(closure, params, tau, *args, sample_size=1, **kwargs):
    with torch.no_grad():
        params_0 = [y.clone() for y in params]
        if sample_size == 1:
            inexact_gradient = zo_grad_central(closure, params, tau)
        else:
            inexact_gradient = [torch.zeros_like(y) for y in params]
            for _ in range(sample_size):
                in_grad = zo_grad_central(closure, params, tau)
                for e, g in zip(in_grad, inexact_gradient):
                    g += e
        for y, p in zip(params, params_0):
            y.zero_().add_(p)
    return inexact_gradient


def zo_grad_forward_coord(closure, params, tau, *args, **kwargs):
    with torch.no_grad():
        f_0 = closure()
        deltas = [torch.zeros_like(y) for y in params]
        for param, delta in zip(params, deltas):
            for y, d in zip(param, delta):
                y.add_(tau)
                d.add_((closure() - f_0) / tau)
    return deltas


def zo_grad_central_coord(closure, params, tau, *args, **kwargs):
    with torch.no_grad():
        deltas = [torch.zeros_like(y) for y in params]
        for param, delta in zip(params, deltas):
            for y, d in zip(param, delta):
                y.add_(tau)
                f_1 = closure()
                y.sub_(2 * tau)
                f_0 = closure()
                y.add_(tau)
                d.add_((f_1 - f_0) / (2 * tau))
    return deltas